@@ -50,6 +50,12 @@ DEFAULT_USER_AGENT = 'keystoneauth1/%s %s %s/%s' % (
 # here and we'll add it to the list as required.
 _LOG_CONTENT_TYPES = set(['application/json'])
 
+_MAX_RETRY_INTERVAL = 60.0
+
+# NOTE(efried): This is defined in oslo_middleware.request_id.INBOUND_HEADER,
+# but it didn't seem worth adding oslo_middleware to requirements just for that
+_REQUEST_ID_HEADER = 'X-Openstack-Request-Id'
+
 
 def _construct_session(session_obj=None):
     # NOTE(morganfainberg): if the logic in this function changes be sure to
@@ -79,6 +85,38 @@ def _mv_legacy_headers_for_service(mv_service_type):
     return headers
 
 
+def _sanitize_headers(headers):
+    """Ensure headers are strings and not bytes."""
+    str_dict = {}
+    for k, v in headers.items():
+        if six.PY3:
+            # requests expects headers to be str type in python3, which means
+            # if we get a bytes we need to decode it into a str
+            k = k.decode('ASCII') if isinstance(k, six.binary_type) else k
+            if v is not None:
+                v = v.decode('ASCII') if isinstance(v, six.binary_type) else v
+        else:
+            # requests expects headers to be str type in python2, which means
+            # if we get a unicode we need to encode it to ASCII into a str
+            k = k.encode('ASCII') if isinstance(k, six.text_type) else k
+            if v is not None:
+                v = v.encode('ASCII') if isinstance(v, six.text_type) else v
+        str_dict[k] = v
+    return str_dict
+
+
+class NoOpSemaphore(object):
+    """Empty context manager for use as a default semaphore."""
+
+    def __enter__(self):
+        """Enter the context manager and do nothing."""
+        pass
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        """Exit the context manager and do nothing."""
+        pass
+
+
 class _JSONEncoder(json.JSONEncoder):
 
     def default(self, o):
@@ -157,7 +195,7 @@ def _determine_calling_package():
 
 
 def _determine_user_agent():
-    """Attempt to programatically generate a user agent string.
+    """Attempt to programmatically generate a user agent string.
 
     First, look at the name of the process. Return this unless it is in
     the `ignored` list.  Otherwise, look at the function call stack and
@@ -183,6 +221,24 @@ def _determine_user_agent():
     return name
 
 
+class RequestTiming(object):
+    """Contains timing information for an HTTP interaction."""
+
+    #: HTTP method used for the call (GET, POST, etc)
+    method = None
+
+    #: URL against which the call was made
+    url = None
+
+    #: Elapsed time information
+    elapsed = None  # type: datetime.timedelta
+
+    def __init__(self, method, url, elapsed):
+        self.method = method
+        self.url = url
+        self.elapsed = elapsed
+
+
 class Session(object):
     """Maintains client communication state and common functionality.
 
@@ -244,6 +300,12 @@ class Session(object):
                                  None which means automatically manage)
     :param bool split_loggers: Split the logging of requests across multiple
                                loggers instead of just one. Defaults to False.
+    :param bool collect_timing: Whether or not to collect per-method timing
+                                information for each API call. (optional,
+                                defaults to False)
+    :param rate_semaphore: Semaphore to be used to control concurrency
+                           and rate limiting of requests. (optional,
+                           defaults to no concurrency or rate control)
     """
 
     user_agent = None
@@ -256,7 +318,8 @@ class Session(object):
                  cert=None, timeout=None, user_agent=None,
                  redirect=_DEFAULT_REDIRECT_LIMIT, additional_headers=None,
                  app_name=None, app_version=None, additional_user_agent=None,
-                 discovery_cache=None, split_loggers=None):
+                 discovery_cache=None, split_loggers=None,
+                 collect_timing=False, rate_semaphore=None):
 
         self.auth = auth
         self.session = _construct_session(session)
@@ -276,6 +339,9 @@ class Session(object):
         # NOTE(mordred) split_loggers kwarg default is None rather than False
         # so we can distinguish between the value being set or not.
         self._split_loggers = split_loggers
+        self._collect_timing = collect_timing
+        self._api_times = []
+        self._rate_semaphore = rate_semaphore or NoOpSemaphore()
 
         if timeout is not None:
             self.timeout = float(timeout)
@@ -321,10 +387,10 @@ class Session(object):
         secure_headers = ('authorization', 'x-auth-token',
                           'x-subject-token', 'x-service-token')
         if header[0].lower() in secure_headers:
-            token_hasher = hashlib.sha1()
+            token_hasher = hashlib.sha256()
             token_hasher.update(header[1].encode('utf-8'))
             token_hash = token_hasher.hexdigest()
-            return (header[0], '{SHA1}%s' % token_hash)
+            return (header[0], '{SHA256}%s' % token_hash)
         return header
 
     def _get_split_loggers(self, split_loggers):
@@ -516,7 +582,8 @@ class Session(object):
                 endpoint_override=None, connect_retries=0, logger=None,
                 allow=None, client_name=None, client_version=None,
                 microversion=None, microversion_service_type=None,
-                **kwargs):
+                status_code_retries=0, retriable_status_codes=None,
+                rate_semaphore=None, global_request_id=None, **kwargs):
         """Send an HTTP request with the specified characteristics.
 
         Wrapper around `requests.Session.request` to handle tasks such as
@@ -594,6 +661,18 @@ class Session(object):
                        provided or does not have a service_type, microversion
                        is given and microversion_service_type is not provided,
                        an exception will be raised.
+        :param int status_code_retries: the maximum number of retries that
+                                        should be attempted for retriable
+                                        HTTP status codes (optional, defaults
+                                        to 0 - never retry).
+        :param list retriable_status_codes: list of HTTP status codes that
+                                            should be retried (optional,
+                                            defaults to HTTP 503, has no effect
+                                            when status_code_retries is 0).
+        :param rate_semaphore: Semaphore to be used to control concurrency
+                               and rate limiting of requests. (optional,
+                               defaults to no concurrency or rate control)
+        :param global_request_id: Value for the X-Openstack-Request-Id header.
         :param kwargs: any other parameter that can be passed to
                        :meth:`requests.Session.request` (such as `headers`).
                        Except:
@@ -615,6 +694,9 @@ class Session(object):
         else:
             split_loggers = None
         logger = logger or utils.get_logger(__name__)
+        # HTTP 503 - Service Unavailable
+        retriable_status_codes = retriable_status_codes or [503]
+        rate_semaphore = rate_semaphore or self._rate_semaphore
 
         headers = kwargs.setdefault('headers', dict())
         if microversion:
@@ -708,9 +790,19 @@ class Session(object):
             headers.setdefault('Content-Type', 'application/json')
             kwargs['data'] = self._json.encode(json)
 
+        if global_request_id is not None:
+            # NOTE(efried): This does *not* setdefault. If a global_request_id
+            # kwarg was explicitly specified, it should override any value
+            # previously configured (e.g. in Adapter.global_request_id).
+            headers[_REQUEST_ID_HEADER] = global_request_id
+
         for k, v in self.additional_headers.items():
             headers.setdefault(k, v)
 
+        # Bug #1766235: some headers may be bytes
+        headers = _sanitize_headers(headers)
+        kwargs['headers'] = headers
+
         kwargs.setdefault('verify', self.verify)
 
         if requests_auth:
@@ -737,7 +829,9 @@ class Session(object):
 
         send = functools.partial(self._send_request,
                                  url, method, redirect, log, logger,
-                                 split_loggers, connect_retries)
+                                 split_loggers, connect_retries,
+                                 status_code_retries, retriable_status_codes,
+                                 rate_semaphore)
 
         try:
             connection_params = self.get_auth_connection_params(auth=auth)
@@ -808,10 +902,26 @@ class Session(object):
                          resp.status_code)
             raise exceptions.from_response(resp, method, url)
 
+        if self._collect_timing:
+            for h in resp.history:
+                self._api_times.append(RequestTiming(
+                    method=h.request.method,
+                    url=h.request.url,
+                    elapsed=h.elapsed,
+                ))
+            self._api_times.append(RequestTiming(
+                method=resp.request.method,
+                url=resp.request.url,
+                elapsed=resp.elapsed,
+            ))
+
         return resp
 
     def _send_request(self, url, method, redirect, log, logger, split_loggers,
-                      connect_retries, connect_retry_delay=0.5, **kwargs):
+                      connect_retries, status_code_retries,
+                      retriable_status_codes, rate_semaphore,
+                      connect_retry_delay=0.5, status_code_retry_delay=0.5,
+                      **kwargs):
         # NOTE(jamielennox): We handle redirection manually because the
         # requests lib follows some browser patterns where it will redirect
         # POSTs as GETs for certain statuses which is not want we want for an
@@ -825,7 +935,8 @@ class Session(object):
 
         try:
             try:
-                resp = self.session.request(method, url, **kwargs)
+                with rate_semaphore:
+                    resp = self.session.request(method, url, **kwargs)
             except requests.exceptions.SSLError as e:
                 msg = 'SSL exception connecting to %(url)s: %(error)s' % {
                     'url': url, 'error': e}
@@ -855,10 +966,16 @@ class Session(object):
                         {'e': e, 'delay': connect_retry_delay})
             time.sleep(connect_retry_delay)
 
+            connect_retry_delay = min(connect_retry_delay * 2,
+                                      _MAX_RETRY_INTERVAL)
+
             return self._send_request(
                 url, method, redirect, log, logger, split_loggers,
+                status_code_retries=status_code_retries,
+                retriable_status_codes=retriable_status_codes,
+                rate_semaphore=rate_semaphore,
                 connect_retries=connect_retries - 1,
-                connect_retry_delay=connect_retry_delay * 2,
+                connect_retry_delay=connect_retry_delay,
                 **kwargs)
 
         if log:
@@ -887,13 +1004,38 @@ class Session(object):
                 # This request actually worked so we can reset the delay count.
                 new_resp = self._send_request(
                     location, method, redirect, log, logger, split_loggers,
+                    rate_semaphore=rate_semaphore,
                     connect_retries=connect_retries,
+                    status_code_retries=status_code_retries,
+                    retriable_status_codes=retriable_status_codes,
                     **kwargs)
 
                 if not isinstance(new_resp.history, list):
                     new_resp.history = list(new_resp.history)
                 new_resp.history.insert(0, resp)
                 resp = new_resp
+        elif (resp.status_code in retriable_status_codes and
+              status_code_retries > 0):
+
+            logger.info('Retriable status code %(code)s. Retrying in '
+                        '%(delay).1fs.',
+                        {'code': resp.status_code,
+                         'delay': status_code_retry_delay})
+            time.sleep(status_code_retry_delay)
+
+            status_code_retry_delay = min(status_code_retry_delay * 2,
+                                          _MAX_RETRY_INTERVAL)
+
+            # NOTE(jamielennox): We don't pass through connect_retry_delay.
+            # This request actually worked so we can reset the delay count.
+            return self._send_request(
+                url, method, redirect, log, logger, split_loggers,
+                connect_retries=connect_retries,
+                status_code_retries=status_code_retries - 1,
+                retriable_status_codes=retriable_status_codes,
+                rate_semaphore=rate_semaphore,
+                status_code_retry_delay=status_code_retry_delay,
+                **kwargs)
 
         return resp
 
@@ -1048,6 +1190,38 @@ class Session(object):
         auth = self._auth_required(auth, 'determine endpoint URL')
         return auth.get_api_major_version(self, **kwargs)
 
+    def get_all_version_data(self, auth=None, interface='public',
+                             region_name=None, service_type=None,
+                             **kwargs):
+        """Get version data for all services in the catalog.
+
+        :param auth:
+            The auth plugin to use for token. Overrides the plugin on
+            the session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+        :param interface:
+            Type of endpoint to get version data for. Can be a single value
+            or a list of values. A value of None indicates that all interfaces
+            should be queried. (optional, defaults to public)
+        :param string region_name:
+            Region of endpoints to get version data for. A valueof None
+            indicates that all regions should be queried. (optional, defaults
+            to None)
+        :param string service_type:
+            Limit the version data to a single service. (optional, defaults
+            to None)
+        :returns: A dictionary keyed by region_name with values containing
+            dictionaries keyed by interface with values being a list of
+            `~keystoneauth1.discover.VersionData`.
+        """
+        auth = self._auth_required(auth, 'determine endpoint URL')
+        return auth.get_all_version_data(
+            self,
+            interface=interface,
+            region_name=region_name,
+            service_type=service_type,
+            **kwargs)
+
     def get_auth_connection_params(self, auth=None, **kwargs):
         """Return auth connection params as provided by the auth plugin.
 
@@ -1148,6 +1322,17 @@ class Session(object):
         auth = self._auth_required(auth, 'get project_id')
         return auth.get_project_id(self)
 
+    def get_timings(self):
+        """Return collected API timing information.
+
+        :returns: List of `RequestTiming` objects.
+        """
+        return self._api_times
+
+    def reset_timings(self):
+        """Clear API timing information."""
+        self._api_times = []
+
 
 REQUESTS_VERSION = tuple(int(v) for v in requests.__version__.split('.'))
 
